/*
 * Copyright (C) 2009, Texas Instruments, Incorporated
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * ATMEL AT93CX6 eeprom supported
 * bus with: x8,
 * note: bus with x16 not supported.
 */
#include <common.h>
#include <asm/arch/sys_proto.h>

#ifdef CONFIG_DVR_EEPROM
#include "gpio.h"

#define EEPROM_CS_ACTIVE		1
#define EEPROM_CS_INACTIVE		0
#define EEPROM_DELAY			10

#define AT93C_WRITE_ON			0x180
#define AT93C_WRITE_OFF			0x0
#define AT93C_ER_ALL			0x100

#define AT93C_CODE				0
#define AT93C_WRITE_CODE		0x1
#define AT93C_READ_CODE			0x2
#define AT93C_ERASE_CODE		0x3

#define mdelay(n)				udelay((n)*1000)

#define ROM_MAC_SYNC			0x4E		//# 'N' for mac is valid
#define ROM_MAC_ADDR0			0
#define ROM_MAC_SIZE			8
#define ROM_MAC_ADDR1			(ROM_MAC_ADDR0+ROM_MAC_SIZE)
#define ROM_HW_ADDR				(ROM_MAC_ADDR1+ROM_MAC_SIZE)
#define ROM_HW_SIZE				2

static void spi_clk(void)
{
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SCLK, 1);
	udelay(EEPROM_DELAY);
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SCLK, 0);
	udelay(EEPROM_DELAY);
}

static u8 spi_r(void)
{
	int i;
	u8 temp = 0;
	
	for (i = 7; i >= 0; i--) {
		gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SCLK, 1);
		udelay(EEPROM_DELAY);

		temp |= 
			(gpio_get_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDI)) ? 1 << i : 0;

		gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SCLK, 0);
		udelay(EEPROM_DELAY);
	}
	
	return temp;
}

static void spi_w(u8 data)
{
	int i;

	for (i = 7; i >= 0; i--) {
		(data & (0x01 << i)) ? gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 1) :
			gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 0);
		spi_clk();
	}
}

static void spi_reg(u8 data, u16 reg)
{
	int i;

	/* Initial start bit */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 1);
	spi_clk();

	/* AT93C operation
	 *(2-bit op code, msb first) 
	 */
	for (i = 1; i >= 0; i--) {
		(data & (0x01 << i)) ? gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 1) :
			gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 0);
		spi_clk();
	}

	/* Address location (9-bit) */
	for (i = 8; i >= 0; i--) {
		(reg & (0x01 << i)) ? gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 1) :
			gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 0);
		spi_clk();
	}
}

/**
 * eeprom_93c6x_read - read from AT93CX6 EEPROM
 * @reg:	The register offset.
 *
 * This function reads a word from the AT93CX6 EEPROM.
 *
 * Return the data value.
 */
static u8 spi_read_word(u16 reg)
{
	u8 data;

	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	spi_reg(AT93C_READ_CODE, (reg&0x1ff));
	data = spi_r();
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);

	return data;
}

/**
 * eeprom_93c6x_write - write to AT93CX6 EEPROM
 * @reg:	The register offset.
 * @data:	The data value.
 *
 * This procedure writes a word(8bit) to the AT93CX6 EEPROM.
 */
static void spi_write_word(u16 reg, u8 data)
{
	int timeout = 0;
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	/* Enable write. */
	spi_reg(AT93C_CODE, AT93C_WRITE_ON);
	/* inactive cs */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Erase the register. */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	spi_reg(AT93C_ERASE_CODE, reg);
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Check operation complete. */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	timeout = 8;
	mdelay(2);
	
	do {
		mdelay(1);
	} while (!gpio_get_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDI) && --timeout);
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Write the register. */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	spi_reg(AT93C_WRITE_CODE, reg);
	spi_w(data);
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Check operation complete. (max 10ms) */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	timeout = 8;
	mdelay(2);
	
	do {
		mdelay(1);
	} while (!gpio_get_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDI) && --timeout);
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Disable write. */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	spi_reg(AT93C_CODE, AT93C_WRITE_OFF);
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
}

static int eeprom_93c6x_read(u32 offset, size_t len, void *buf)
{
	u8 *p, *tmp;
	unsigned int i;

	p = tmp = (u8 *)buf;
	
	if (offset >= CONFIG_SYS_EEPROM_SIZE)
		return 0;
	
	if ((offset + len) > CONFIG_SYS_EEPROM_SIZE) {
		len = (CONFIG_SYS_EEPROM_SIZE - offset);
	}
	
	for (i = offset; len > 0; i++, --len) 
		*p++ = spi_read_word(i);

	return (int)(p-tmp);
}

static int eeprom_93c6x_write(u32 offset, size_t len, const void *buf)
{
	u32 i;
	u8 *p, *tmp;

	p = tmp = (u8 *)buf;

	if (offset >= CONFIG_SYS_EEPROM_SIZE)
		return 0;
	
	if ((offset + len) > CONFIG_SYS_EEPROM_SIZE) {
		len = (CONFIG_SYS_EEPROM_SIZE - offset);
	}
	
	for (i = offset; len > 0; ++i, --len) {
		spi_write_word((i&0x1ff), *p++);
	}

	return (int)(p-tmp);
}

static int eeprom_93c6x_erase(void)
{
	int timeout = 0;
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	/* Enable write. */
	spi_reg(AT93C_CODE, AT93C_WRITE_ON);
	/* inactive cs */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	/* erase all */
	spi_reg(AT93C_CODE, AT93C_ER_ALL);
	/* inactive cs */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	/* Check operation complete. (max 10ms) */
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_ACTIVE);
	timeout = 8;
	mdelay(2);
	
	do {
		mdelay(1);
	} while (!gpio_get_value(GPIO_BANK0, CONFIG_SYS_GPIO_SDI) && --timeout);
	
	gpio_set_value(GPIO_BANK0, CONFIG_SYS_GPIO_CS, EEPROM_CS_INACTIVE);
	udelay(1);
	
	return 0;
}

/******************************************************************************
 * DVR Board specific command saving to eeprom
 *****************************************************************************/
/*
	ROM area
	SYNC  MAC_ADDR DUMMY
	1Byte 6Byte    1Byte
	H/W Ver
	2Byte
*/
int dvr_eeprom_init(void)
{
	//# enable GPIO for SPI - default '0'
	gpio_direction_output(GPIO_BANK0, CONFIG_SYS_GPIO_CS, 0);
	gpio_direction_output(GPIO_BANK0, CONFIG_SYS_GPIO_SCLK, 0);
	gpio_direction_output(GPIO_BANK0, CONFIG_SYS_GPIO_SDO, 0);
	
	gpio_direction_input(GPIO_BANK0, CONFIG_SYS_GPIO_SDI);

	return 0;
}

int dvr_get_hw_version(void)
{
	unsigned char buf[ROM_HW_SIZE];
	int dvr_hwver=0;

	eeprom_93c6x_read(ROM_HW_ADDR, ROM_HW_SIZE, (void *)buf);
	dvr_hwver = (buf[0]<<8) + buf[1];

	return dvr_hwver;
}

int dvr_eeprom_read(char *buf, int id)
{
	char tmp[ROM_MAC_SIZE];
	int addr;

	if (id == 0) {
		//# read mac addr
		addr = ROM_MAC_ADDR0;
		/* Set Ethernet MAC address from SPI EEPROM */
		if (!eeprom_93c6x_read(addr, ROM_MAC_SIZE, (void *)tmp)) {
			printf("\nEEPROM read FAILED!!!\n");
		} 
		else {
			if (tmp[0] == ROM_MAC_SYNC) {
				sprintf((char *)buf, "%02x:%02x:%02x:%02x:%02x:%02x",
						tmp[1], tmp[2], tmp[3], tmp[4], tmp[5], tmp[6]);
				return 1;
			}
		}
	}
	else if (id == 2) {
		//# read hw version
		addr = ROM_HW_ADDR;
		if (!eeprom_93c6x_read(addr, ROM_HW_SIZE, (void *)tmp)) {
			printf("\nEEPROM read FAILED!!!\n");
		} else {
			sprintf(buf, "%02x:%02x", tmp[0], tmp[1]);
			return 1;
		}
	}

	return 0;
}

int dvr_eeprom_write(char *buf, int id)
{
	if (id == 0) {
		buf[0] = ROM_MAC_SYNC;
		buf[7] = 0;
		
		eeprom_93c6x_write(ROM_MAC_ADDR0, ROM_MAC_SIZE, (const void *)buf);
	}
	else if (id == 2) {
		eeprom_93c6x_write(ROM_HW_ADDR, ROM_HW_SIZE, (const void *)buf);
	}

	return 0;
}

static int do_save_dvrrom (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
	int i;
	
	u_int8_t env_ethaddr[ROM_MAC_SIZE];
	u_int8_t env_ver[2];
	char *s = NULL,*end;

	/* at least two arguments please */
	if (argc < 2)
		goto usage;
	
	if (strcmp (argv[1], "save") == 0) {
		/* Read Ethernet MAC address from the U-Boot environment. */
		s = getenv("ethaddr");
		
		if(s == NULL) {
			printf ("No valid MAC address in U-boot environment\n");
		} else {
			env_ethaddr[0] = ROM_MAC_SYNC;
			for (i = 1; i < ROM_MAC_SIZE-1; i++) {
				env_ethaddr[i] = s ? simple_strtoul(s, &end, 16) : 0;
				if (s)
					s = (*end) ? end+1 : end;
			}
			env_ethaddr[7] = 0;

			eeprom_93c6x_write(ROM_MAC_ADDR0, ROM_MAC_SIZE, (const void *)env_ethaddr);
			printf("Set MAC address0 from environment\n");
		}
		
		/* Read Ethernet1 MAC address from the U-Boot environment. */
		s = getenv("eth1addr");
		if(s == NULL) {
			printf ("No valid MAC address1 in U-boot environment\n");
		} else {
			env_ethaddr[0] = ROM_MAC_SYNC;
			for (i = 1; i < ROM_MAC_SIZE-1; i++) {
				env_ethaddr[i] = s ? simple_strtoul(s, &end, 16) : 0;
				if (s)
					s = (*end) ? end+1 : end;
			}
			env_ethaddr[7] = 0;
			
			eeprom_93c6x_write(ROM_MAC_ADDR1, ROM_MAC_SIZE, (const void *)env_ethaddr);
			printf("Set MAC address1 from environment\n");
		}

		/* Read H/W Version from the U-Boot environment. */
		s = getenv("hwver");
		if (s == NULL) {
			printf ("No valid H/W Version in U-boot environment\n");
		} 
		else {
			for (i = 0; i < 2; i++) {
				env_ver[i] = s ? simple_strtoul(s, &end, 16) : 0;
				if (s)
					s = (*end) ? end+1 : end;
			}

			eeprom_93c6x_write(ROM_HW_ADDR, ROM_HW_SIZE, (const void *)env_ver);
			printf("Set H/W Version(%2x.%02x)from environment\n", env_ver[0], env_ver[1]);
		}
	}
	else if (strcmp (argv[1], "write") == 0) {
		printf ("do not support now\n");
	} 
	else if (strcmp (argv[1], "read") == 0) {
		/* at least three arguments please */
		if (argc < 3)
			goto usage;
		
		if (strcmp (argv[2], "ethaddr") == 0) {
			eeprom_93c6x_read(ROM_MAC_ADDR0, ROM_MAC_SIZE, (char *)env_ethaddr);
			printf("%02x:%02x:%02x:%02x:%02x:%02x\n",
				env_ethaddr[1], env_ethaddr[2], env_ethaddr[3],
				env_ethaddr[4], env_ethaddr[5], env_ethaddr[6]);
		}
		else if (strcmp (argv[2], "eth1addr") == 0) {
			eeprom_93c6x_read(ROM_MAC_ADDR1, ROM_MAC_SIZE, (char *)env_ethaddr);
			printf("%02x:%02x:%02x:%02x:%02x:%02x\n",
				env_ethaddr[1], env_ethaddr[2], env_ethaddr[3],
				env_ethaddr[4], env_ethaddr[5], env_ethaddr[6]);
		}
		else if (strcmp (argv[2], "hwver") == 0) {
			eeprom_93c6x_read(ROM_HW_ADDR, ROM_HW_SIZE, (char *)env_ethaddr);
			printf("%02x:%02x\n", env_ethaddr[0], env_ethaddr[1]);
		} 
		else {
			printf("Invalid arguments\n");
			goto usage;
		}
	}
	else if (strcmp (argv[1], "eral") == 0) {
		eeprom_93c6x_erase();
		printf ("erase all\n");
	} 
	else {
		goto usage;
	}

	return 0;

usage:
	cmd_usage(cmdtp);
	return 1;
}

U_BOOT_CMD(
	dvrrom, 3, 1,	do_save_dvrrom,
	"Saving environment to eeprom",
	"save - saveing ethaddr, hwver\n"
	"dvrrom read [ethaddr/eth1addr/hwver]\n"
	"dvrrom eral - erase all rom\n"
);
#endif /* 3-wire eeprom support */

